##For GNU GLP v3:
##
##Coordination of Power and Natural Gas Systems: Convexification Approaches of Linepack Modeling
##Copyright (C) 2018 Anna Schwele
##
##This program is free software: you can redistribute it and/or modify
##t under the terms of the GNU General Public License as published by
##the Free Software Foundation, either version 3 of the License, or
##(at your option) any later version.
##
##This program is distributed in the hope that it will be useful,
##but WITHOUT ANY WARRANTY; without even the implied warranty of
##MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##GNU General Public License for more details.

import numpy as np
import pandas as pd
import gurobipy as gb
import pprint

def input():

	solution = {

	'PP':{
	'i1':{'Pmax':152,'ELnode':'n1', 'phi':12.65, 'NGnode':'m12','type':'GFPP'},
	'i2':{'Pmax':152,'ELnode':'n2', 'phi':20.25, 'NGnode':'m12','type':'GFPP'},
	'i3':{'Pmax':300,'ELnode':'n7', 'C_E':65.61,'type':'TPP'},
	'i4':{'Pmax':591,'ELnode':'n13', 'C_E':30.82,'type':'TPP'},
	'i5':{'Pmax':60,'ELnode':'n15', 'phi':11.12, 'NGnode':'m10','type':'GFPP'},
	'i6':{'Pmax':155,'ELnode':'n15', 'phi':15.1, 'NGnode':'m10','type':'GFPP'},
	'i7':{'Pmax':155,'ELnode':'n16', 'phi':14.88, 'NGnode':'m7','type':'GFPP'},
	'i8':{'Pmax':400,'ELnode':'n18', 'C_E':20.84,'type':'TPP'},
	'i9':{'Pmax':400,'ELnode':'n21', 'C_E':26.9,'type':'TPP'},
	'i10':{'Pmax':300,'ELnode': 'n22', 'phi':13.3, 'NGnode':'m6','type':'GFPP'},
	'i11':{'Pmax':60,'ELnode': 'n23', 'phi':16.8, 'NGnode':'m6','type':'GFPP'},
	'i12':{'Pmax':350,'ELnode':'n23', 'C_E':32.22,'type':'TPP'}},

	'Windfarm':{
	'j1':{'Wmax':500, 'ELnode': 'n5'},
	'j2':{'Wmax':1000, 'ELnode': 'n7'}},

	'WindFactor':{
	't1':{'j1':0.8,'j2':0.9},
	't2':{'j1':0.9,'j2':0.85},
	't3': {'j1':0.85,'j2':0.75},
	't4': {'j1':0.6,'j2':0.7},
	't5': {'j1':0.7,'j2':0.8},
	't6':{'j1':0.68,'j2':0.58},
	't7': {'j1':0.5,'j2':0.56},
	't8': {'j1':0.3,'j2':0.4},
	't9': {'j1':0.45,'j2':0.2},
	't10':{'j1':0.85,'j2':0.75},
	't11': {'j1':0.7,'j2':0.75},
	't12':{'j1':0.15,'j2':0.35},
	't13': {'j1':0.2,'j2':0.25},
	't14':{'j1':0.4,'j2':0.3},
	't15': {'j1':0.35,'j2':0.25},
	't16':{'j1':0.65,'j2':0.7},
	't17': {'j1':0.5,'j2':0.2},
	't18':{'j1':0.25,'j2':0.1},
	't19': {'j1':0.1,'j2':0.15},
	't20': {'j1':0.2,'j2':0.15},
	't21': {'j1':0.15,'j2':0.1},
	't22': {'j1':0.15,'j2':0.1},
	't23': {'j1':0.5,'j2':0.7},
	't24': {'j1':0.65,'j2':0.77}},

	'GasSupply':{
	'k1':{'Gmax':6000, 'C_G':2, 'NGnode':'m1'},
	'k2':{'Gmax':6000, 'C_G':2.4, 'NGnode':'m3'},
	'k3':{'Gmax':15000, 'C_G':3.2, 'NGnode':'m11'}},

	'ELDemand':{
	'd1' : {'share' : 0.038,'node': 'n1'},
	'd2' : {'share' : 0.034,'node': 'n2'},
	'd3' : {'share' : 0.063,'node': 'n3'},
	'd4' : {'share' : 0.026,'node': 'n4'},
	'd5' : {'share' : 0.025,'node': 'n5'},
	'd6' : {'share' : 0.048,'node': 'n6'},
	'd7' : {'share' : 0.044,'node': 'n7'},
	'd8' : {'share' : 0.060,'node': 'n8'},
	'd9' : {'share' : 0.061,'node': 'n9'},
	'd10' : {'share' : 0.068,'node': 'n10'},
	'd11' : {'share' : 0.093,'node': 'n13'},
	'd12' : {'share' : 0.068,'node': 'n14'},
	'd13' : {'share' : 0.111,'node': 'n15'},
	'd14' : {'share' : 0.035,'node': 'n16'},
	'd15' : {'share' : 0.117,'node': 'n18'},
	'd16' : {'share' : 0.064,'node': 'n19'},
	'd17' : {'share' : 0.045,'node': 'n20'}},

	'NGDemand':{
	'd1' : {'share' : 0.25,'node': 'm5'},
	'd2' : {'share' : 0.25,'node': 'm7'},
	'd3' : {'share' : 0.35,'node': 'm6'},
	'd4' : {'share' : 0.15,'node': 'm12'}},

	'Demand':{
	't1':{'EL':2108.73, 'NG':7000},
	't2':{'EL':2002.63, 'NG':6700,'prev':'t1'},
	't3':{'EL':1573.98, 'NG':6400,'prev':'t2'},
	't4':{'EL':1230.90, 'NG':6500,'prev':'t3'},
	't5':{'EL':1251.00, 'NG':6600,'prev':'t4'},
	't6':{'EL':1282.04, 'NG':6700,'prev':'t5'},
	't7':{'EL':1778.60, 'NG':7750,'prev':'t6'},
	't8':{'EL':2344.03, 'NG':7800,'prev':'t7'},
	't9':{'EL':2461.77, 'NG':8550,'prev':'t8'},
	't10':{'EL':2702.58, 'NG':8700,'prev':'t9'},
	't11':{'EL':2741.14, 'NG':8700,'prev':'t10'},
	't12':{'EL':2535.38, 'NG':8550,'prev':'t11'},
	't13':{'EL':2482.55, 'NG':8550,'prev':'t12'},
	't14':{'EL':2367.35, 'NG':8550,'prev':'t13'},
	't15':{'EL':2741.30, 'NG':8400,'prev':'t14'},
	't16':{'EL':2875.31, 'NG':8550,'prev':'t15'},
	't17':{'EL':2597.63, 'NG':9000,'prev':'t16'},
	't18':{'EL':2715.35, 'NG':9000,'prev':'t17'},
	't19':{'EL':2879.81, 'NG':9000,'prev':'t18'},
	't20':{'EL':2989.47, 'NG':8700,'prev':'t19'},
	't21':{'EL':3000.00, 'NG':8250,'prev':'t20'},
	't22':{'EL':2713.75, 'NG':7500,'prev':'t21'},
	't23':{'EL':2610.42, 'NG':6600,'prev':'t22'},
	't24':{'EL':2437.59, 'NG':6700,'prev':'t23'}},

	'Lines': {
	'l1' : {'From' : 'n1', 'To' : 'n2','B' : 0.0146, 'capacity' : 175},
	'l2' : {'From' : 'n1', 'To' : 'n3','B' : 0.2253, 'capacity' : 175},
	'l3' : {'From' : 'n1', 'To' : 'n5','B' : 0.0907, 'capacity' : 500},
	'l4' : {'From' : 'n2', 'To' : 'n4','B' : 0.1356, 'capacity' : 175},
	'l5' : {'From' : 'n2', 'To' : 'n6','B' : 0.205, 'capacity' : 175},
	'l6' : {'From' : 'n3', 'To' : 'n9','B' : 0.1271, 'capacity' : 175},
	'l7' : {'From' : 'n3', 'To' : 'n24','B' : 0.084, 'capacity' : 400},
	'l8' : {'From' : 'n4', 'To' : 'n9','B' : 0.111, 'capacity' : 175},
	'l9' : {'From' : 'n5', 'To' : 'n10','B' : 0.094, 'capacity' : 500},
	'l10' : {'From' : 'n6', 'To' : 'n10','B' : 0.0642, 'capacity' : 175},
	'l11' : {'From' : 'n7', 'To' : 'n8','B' : 0.0652, 'capacity' : 1000},
	'l12' : {'From' : 'n8', 'To' : 'n9','B' : 0.1762, 'capacity' : 175},
	'l13' : {'From' : 'n8', 'To' : 'n10','B' : 0.1762, 'capacity' : 175},
	'l14' : {'From' : 'n9', 'To' : 'n11','B' : 0.084, 'capacity' : 400},
	'l15' : {'From' : 'n9', 'To' : 'n12','B' : 0.084, 'capacity' : 400},
	'l16' : {'From' : 'n10', 'To' : 'n11','B' : 0.084, 'capacity' : 400},
	'l17' : {'From' : 'n10', 'To' : 'n12','B' : 0.084, 'capacity' : 400},
	'l18' : {'From' : 'n11', 'To' : 'n13','B' : 0.0488, 'capacity' : 500},
	'l19' : {'From' : 'n11', 'To' : 'n14','B' : 0.0426, 'capacity' : 500},
	'l20' : {'From' : 'n12', 'To' : 'n13','B' : 0.0488, 'capacity' : 500},
	'l21' : {'From' : 'n12', 'To' : 'n23','B' : 0.0985, 'capacity' : 500},
	'l22' : {'From' : 'n13', 'To' : 'n23','B' : 0.0884, 'capacity' : 500},
	'l23' : {'From' : 'n14', 'To' : 'n16','B' : 0.0594, 'capacity' : 500},
	'l24' : {'From' : 'n15', 'To' : 'n16','B' : 0.0172, 'capacity' : 500},
	'l25' : {'From' : 'n15', 'To' : 'n21','B' : 0.0249, 'capacity' : 500},
	'l26' : {'From' : 'n15', 'To' : 'n24','B' : 0.0529, 'capacity' : 500},
	'l27' : {'From' : 'n16', 'To' : 'n17','B' : 0.0263, 'capacity' : 500},
	'l28' : {'From' : 'n16', 'To' : 'n19','B' : 0.0234, 'capacity' : 500},
	'l29' : {'From' : 'n17', 'To' : 'n18','B' : 0.0143, 'capacity' : 500},
	'l30' : {'From' : 'n17', 'To' : 'n22','B' : 0.1069, 'capacity' : 500},
	'l31' : {'From' : 'n18', 'To' : 'n21','B' : 0.0132, 'capacity' : 1000},
	'l32' : {'From' : 'n19', 'To' : 'n20','B' : 0.0203, 'capacity' : 1000},
	'l33' : {'From' : 'n20', 'To' : 'n23','B' : 0.0112, 'capacity' : 1000},
	'l34' : {'From' : 'n21', 'To' : 'n22','B' : 0.0692, 'capacity' : 500}},

	'Pipelines': {
	'z1':{'From':'m1','To':'m2','K_f':28,'type':'passive','Gamma':1,'K_h':121,'H_ini':39300},
	'z2':{'From':'m2','To':'m4','K_f':28,'type':'passive','Gamma':1.2,'K_h':121,'H_ini':39300},
	'z3':{'From':'m3','To':'m5','K_f':28,'type':'passive','Gamma':1,'K_h':150,'H_ini':49300},
	'z4':{'From':'m4','To':'m5','K_f':21,'type':'passive','Gamma':1,'K_h':186,'H_ini':59300},
	'z5':{'From':'m5','To':'m6','K_f':21,'type':'passive','Gamma':1,'K_h':189,'H_ini':54300},
	'z6':{'From':'m4','To':'m7','K_f':21,'type':'passive','Gamma':1,'K_h':184,'H_ini':54300},
	'z7':{'From':'m6','To':'m8','K_f':28,'type':'passive','Gamma':1,'K_h':150,'H_ini':39300},
	'z8':{'From':'m7','To':'m8','K_f':21,'type':'passive','Gamma':1,'K_h':179,'H_ini':44300},
	'z9':{'From':'m8','To':'m9','K_f':28,'type':'passive','Gamma':1.3,'K_h':148,'H_ini':39300},
	'z10':{'From':'m9','To':'m10','K_f':28,'type':'passive','Gamma':1,'K_h':148,'H_ini':39300},
	'z11':{'From':'m10','To':'m11','K_f':28,'type':'passive','Gamma':1,'K_h':150,'H_ini':39300},
	'z12':{'From':'m11','To':'m12','K_f':28,'type':'passive','Gamma':1,'K_h':130,'H_ini':29300}},

	'ELNodes' : {'n1','n2','n3','n4','n5','n6','n7','n8','n9','n10','n11','n12','n13','n14','n15','n16','n17','n18','n19','n20','n21','n22','n23','n24'},

	'NGNodes':{'m1','m2','m3','m4','m5','m6','m7','m8','m9','m10','m11','m12'},

	'NGPr':{
	'm1':{'minPr': 350, 'maxPr': 470},
	'm2':{'minPr': 310, 'maxPr': 430},
	'm3':{'minPr': 350, 'maxPr': 470},
	'm4':{'minPr': 330, 'maxPr': 460},
	'm5':{'minPr': 290, 'maxPr': 430},
	'm6':{'minPr': 250, 'maxPr': 390},
	'm7':{'minPr': 240, 'maxPr': 360},
	'm8':{'minPr': 230, 'maxPr': 350},
	'm9':{'minPr': 260, 'maxPr': 400},
	'm10':{'minPr': 240, 'maxPr': 360},
	'm11':{'minPr': 250, 'maxPr': 380},
	'm12':{'minPr': 100, 'maxPr': 350}},

	'BigM' : 10000,

	'fixPR':{
	'v1':{'m1': 350, 'm2': 310, 'm3': 350, 'm4': 330, 'm5': 290, 'm6': 250, 'm7': 240, 'm8': 230, 'm9': 260, 'm10': 240, 'm11': 250, 'm12': 100},
	'v2':{'m1': 410, 'm2': 370, 'm3': 410, 'm4': 395, 'm5': 360, 'm6': 320, 'm7': 300, 'm8': 290, 'm9': 330, 'm10': 300, 'm11': 315, 'm12': 225},
	'v3':{'m1': 470, 'm2': 430, 'm3': 470, 'm4': 460, 'm5': 430, 'm6': 290, 'm7': 360, 'm8': 350, 'm9': 400, 'm10': 360, 'm11': 380, 'm12': 350},
	'v4':{'m1': 350, 'm2': 370, 'm3': 350, 'm4': 330, 'm5': 360, 'm6': 250, 'm7': 300, 'm8': 230, 'm9': 330, 'm10': 240, 'm11': 315, 'm12': 100},
	'v5':{'m1': 410, 'm2': 310, 'm3': 410, 'm4': 395, 'm5': 290, 'm6': 320, 'm7': 240, 'm8': 290, 'm9': 260, 'm10': 300, 'm11': 250, 'm12': 225},
	'v6':{'m1': 470, 'm2': 310, 'm3': 470, 'm4': 460, 'm5': 290, 'm6': 390, 'm7': 240, 'm8': 350, 'm9': 260, 'm10': 360, 'm11': 250, 'm12': 350},
	'v7':{'m1': 350, 'm2': 430, 'm3': 350, 'm4': 330, 'm5': 430, 'm6': 250, 'm7': 360, 'm8': 230, 'm9': 400, 'm10': 240, 'm11': 380, 'm12': 100},
	'v8':{'m1': 410, 'm2': 430, 'm3': 410, 'm4': 395, 'm5': 430, 'm6': 320, 'm7': 360, 'm8': 290, 'm9': 400, 'm10': 300, 'm11': 380, 'm12': 225},
	'v9':{'m1': 370, 'm2': 370, 'm3': 470, 'm4': 460, 'm5': 360, 'm6': 390, 'm7': 300, 'm8': 350, 'm9': 330, 'm10': 360, 'm11': 315, 'm12': 350}},

	'KI_plus':{
	'z1':{'From':'m1','To':'m2','v1':60.315,'v2':64.993,'v3':69.359,'v4':0,'v5':42.783,'v6':37.252,'v7':0,'v8':0,'v9':45.406},
	'z2':{'From':'m2','To':'m3','v1':0,'v2':0,'v3':0,'v4':0,'v5':0,'v6':0,'v7':0,'v8':0,'v9':0},
	'z3':{'From':'m3','To':'m5','v1':50.010,'v2':58.507,'v3':69.359,'v4':0,'v5':39.610,'v6':35.581,'v7':0,'v8':0,'v9':43.553},
	'z4':{'From':'m4','To':'m5','v1':44.006,'v2':51.028,'v3':59.118,'v4':0,'v5':30.930,'v6':27.053,'v7':0,'v8':0,'v9':33.734},
	'z5':{'From':'m5','To':'m6','v1':41.437,'v2':45.839,'v3':49.860,'v4':29.185,'v5':0,'v6':0,'v7':25.811,'v8':31.438,'v9':0},
	'z6':{'From':'m4','To':'m7','v1':30.597,'v2':32.282,'v3':33.734,'v4':50.408,'v5':26.440,'v6':24.616,'v7':0,'v8':51.028,'v9':27.702},
	'z7':{'From':'m6','To':'m7','v1':71.443,'v2':66.234,'v3':63.471,'v4':71.443,'v5':66.234,'v6':63.471,'v7':71.443,'v8':66.234,'v9':63.471},
	'z8':{'From':'m7','To':'m8','v1':73.516,'v2':82.019,'v3':89.721,'v4':32.708,'v5':0,'v6':0,'v7':27.298,'v8':35.442,'v9':0},
	'z9':{'From':'m8','To':'m9','v1':0,'v2':0,'v3':0,'v4':0,'v5':0,'v6':0,'v7':0,'v8':0,'v9':0},
	'z10':{'From':'m9','To':'m10','v1':72.800,'v2':67.211,'v3':64.236,'v4':40.796,'v5':0,'v6':0,'v7':35.000,'v8':42.332,'v9':0},
	'z11':{'From':'m10','To':'m11','v1':0,'v2':0,'v3':0,'v4':0,'v5':50.654,'v6':38.913,'v7':0,'v8':0,'v9':57.837},
	'z12':{'From':'m11','To':'m12','v1':30.551,'v2':40.008,'v3':71.898,'v4':29.527,'v5':64.236,'v6':0,'v7':29.023,'v8':34.745,'v9':0}},

	'KO_plus':{
	'z1':{'From':'m1','To':'m2','v1':53.422,'v2':58.652,'v3':63.456,'v4':0,'v5':32.348,'v6':24.570,'v7':0,'v8':0,'v9':35.745},
	'z2':{'From':'m2','To':'m3','v1':0,'v2':0,'v3':0,'v4':0,'v5':0,'v6':0,'v7':0,'v8':0,'v9':0},
	'z3':{'From':'m3','To':'m5','v1':41.437,'v2':51.372,'v3':63.456,'v4':0,'v5':28.017,'v6':21.954,'v7':0,'v8':0,'v9':33.360},
	'z4':{'From':'m4','To':'m5','v1':38.672,'v2':46.507,'v3':55.263,'v4':0,'v5':22.708,'v6':17.055,'v7':0,'v8':0,'v9':26.401},
	'z5':{'From':'m5','To':'m6','v1':35.722,'v2':40.746,'v3':45.222,'v4':20.267,'v5':0,'v6':0,'v7':15.006,'v8':23.396,'v9':0},
	'z6':{'From':'m4','To':'m7','v1':22.252,'v2':24.518,'v3':26.401,'v4':45.826,'v5':16.065,'v6':12.843,'v7':0,'v8':46.507,'v9':18.066},
	'z7':{'From':'m6','To':'m7','v1':0,'v2':60.025,'v3':56.961,'v4':65.728,'v5':60.025,'v6':56.961,'v7':65.728,'v8':60.025,'v9':56.961},
	'z8':{'From':'m7','To':'m8','v1':70.453,'v2':79.285,'v3':87.228,'v4':25.076,'v7':17.440,'v8':28.550,'v9':0},
	'z9':{'From':'m8','To':'m9','v1':0,'v2':0,'v3':0,'v4':0,'v5':0,'v6':0,'v7':0,'v8':0,'v9':0},
	'z10':{'From':'m9','To':'m10','v1':67.200,'v2':61.101,'v3':57.813,'v4':29.670,'v5':0,'v6':0,'v7':21.000,'v8':31.749,'v9':0},
	'z11':{'From':'m10','To':'m11','v1':0,'v2':0,'v3':0,'v4':0,'v5':42.212,'v6':27.023,'v7':0,'v8':0,'v9':50.607},
	'z12':{'From':'m11','To':'m12','v1':12.220,'v2':28.577,'v3':66.222,'v4':9.374,'v5':57.813,'v6':0,'v7':7.638,'v8':20.573,'v9':0}},

	'KI_minus':{
	'z1':{'From':'m1','To':'m2','v1':0,'v2':0,'v3':0,'v4':86.333,'v5':0,'v6':0,'v7':48.199,'v8':92.891,'v9':0},
	'z2':{'From':'m2','To':'m3','v1':0,'v2':0,'v3':0,'v4':0,'v5':0,'v6':0,'v7':0,'v8':0,'v9':0},
	'z3':{'From':'m3','To':'m5','v1':0,'v2':0,'v3':0,'v4':119.628,'v5':0,'v6':0,'v7':48.199,'v8':92.891,'v9':0},
	'z4':{'From':'m4','To':'m5','v1':0,'v2':0,'v3':0,'v4':52.546,'v5':0,'v6':0,'v7':32.755,'v8':53.141,'v9':0},
	'z5':{'From':'m5','To':'m6','v1':0,'v2':0,'v3':0,'v4':0,'v5':49.676,'v6':31.407,'v7':0,'v8':0,'v9':54.600},
	'z6':{'From':'m4','To':'m7','v1':0,'v2':0,'v3':0,'v4':0,'v5':0,'v6':0,'v7':52.546,'v8':0,'v9':0},
	'z7':{'From':'m6','To':'m7','v1':0,'v2':0,'v3':0,'v4':0,'v5':0,'v6':0,'v7':0,'v8':0,'v9':0},
	'z8':{'From':'m7','To':'m8','v1':0,'v2':0,'v3':0,'v4':0,'v5':37.411,'v6':28.851,'v7':0,'v8':0,'v9':40.770},
	'z9':{'From':'m8','To':'m9','v1':0,'v2':0,'v3':0,'v4':0,'v5':0,'v6':0,'v7':0,'v8':0,'v9':0},
	'z10':{'From':'m9','To':'m10','v1':0,'v2':0,'v3':0,'v4':0,'v5':56.125,'v6':40.482,'v7':0,'v8':0,'v9':70.061},
	'z11':{'From':'m10','To':'m11','v1':100.000,'v2':91.830,'v3':87.460,'v4':43.231,'v5':0,'v6':75.440,'v7':36.115,'v8':45.619,'v9':0},
	'z12':{'From':'m11','To':'m12','v1':0,'v2':0,'v3':0,'v4':0,'v5':0,'v6':40.008,'v7':0,'v8':0,'v9':64.236}},

	'KO_minus':{
	'z1':{'From':'m1','To':'m2','v1':0,'v2':0,'v3':0,'v4':81.667,'v5':0,'v6':0,'v7':39.231,'v8':88.570,'v9':0},
	'z2':{'From':'m2','To':'m3','v1':0,'v2':0,'v3':0,'v4':0,'v5':0,'v6':0,'v7':0,'v8':0,'v9':0},
	'z3':{'From':'m3','To':'m5','v1':0,'v2':0,'v3':116.305,'v4':0,'v5':0,'v6':0,'v7':39.231,'v8':88.570,'v9':0},
	'z4':{'From':'m4','To':'m5','v1':0,'v2':0,'v3':0,'v4':48.167,'v5':0,'v6':0,'v7':25.138,'v8':48.815,'v9':0},
	'z5':{'From':'m5','To':'m6','v1':0,'v2':0,'v3':0,'v4':0,'v5':45.019,'v6':23.354,'v7':0,'v8':0,'v9':50.400},
	'z6':{'From':'m4','To':'m7','v1':0,'v2':0,'v3':0,'v4':0,'v5':0,'v6':0,'v7':48.167,'v8':0,'v9':0},
	'z7':{'From':'m6','To':'m7','v1':0,'v2':0,'v3':0,'v4':0,'v5':0,'v6':0,'v7':0,'v8':0,'v9':0},
	'z8':{'From':'m7','To':'m8','v1':0,'v2':0,'v3':0,'v4':0,'v5':30.960,'v6':19.784,'v7':0,'v8':0,'v9':34.946},
	'z9':{'From':'m8','To':'m9','v1':0,'v2':0,'v3':0,'v4':0,'v5':0,'v6':0,'v7':0,'v8':0,'v9':0},
	'z10':{'From':'m9','To':'m10','v1':0,'v2':0,'v3':0,'v4':0,'v5':48.642,'v6':29.237,'v7':0,'v8':0,'v9':64.222},
	'z11':{'From':'m10','To':'m11','v1':96.000,'v2':87.457,'v3':82.857,'v4':32.938,'v5':0,'v6':0,'v7':22.809,'v8':36.015,'v9':0},
	'z12':{'From':'m11','To':'m12','v1':0,'v2':0,'v3':0,'v4':0,'v5':0,'v6':28.577,'v7':0,'v8':0,'v9':57.813}},
	}

	return solution

def del_keys(data, key):
	# Delete column from  dictionary
	return {a:{c:d for c, d in b.items() if c != key} for a, b in data.items()}

def units_in_node(data,key):
	# Mapping
	temp_dict= {}
	for k, v in data.items():
		   for k2, v2 in v.items():
				      temp_dict[(k, k2)] = v2
	length = sum(value == key for value in temp_dict.values())
	solution=[]

	for g in range(length):
			solution.append([k for k, v in temp_dict.items() if v == key][g][0])

	return solution

# Model
def MILP(data):

	# Create a new model
	m = gb.Model()

	# Create indices
	I = [i for i, i_info in data['PP'].items()]
	C = [i for i, i_info in data['PP'].items() if i_info['type']=='TPP']
	G = [i for i, i_info in data['PP'].items() if i_info['type']=='GFPP']
	J = [j for j, j_info in data['Windfarm'].items()]
	T = [t for t, t_info in data['Demand'].items()]
	L = [l for l, l_info in data['Lines'].items()]
	N = list(data['ELNodes'])
	K = [k for k, k_info in data['GasSupply'].items()]
	Z = [z for z, z_info in data['Pipelines'].items()]
	MZ = [mz for mz, mz_info in data['NGPr'].items()]
	R_E = [d for d, d_info in data['ELDemand'].items()]
	R_G = [d for d, d_info in data['NGDemand'].items()]
	V = [v for v, v_info in data['fixPR'].items()]

	l_r = del_keys(data['Lines'],'From')                 #lines ending at nodes
	l_n = del_keys(data['Lines'],'To')                   #lines starting from nodes

	z_m = del_keys(data['Pipelines'],'To')                   #pipelines starting from nodes
	z_u = del_keys(data['Pipelines'],'From')                 #pipelines ending at nodes

	previous = del_keys(data['Demand'],'prev')                 #previous time period

	# Create variables
	p = {}
	for i in I:
		for t in T:
			p[(i,t)] = m.addVar(lb=0.0, ub=data['PP'][i]['Pmax'], name='Generation of power plant; {}; {}; '.format(i,t))
	p

	w = {}
	for j in J:
		for t in T:
			w[(j,t)] = m.addVar(lb=0.0, ub=data['Windfarm'][j]['Wmax']*data['WindFactor'][t][j], name='Generation of wind farm; {}; {}; '.format(j,t))
	w

	g= {}
	for k in K:
		for t in T:
			g[(k,t)] = m.addVar(lb=0.0, ub=data['GasSupply'][k]['Gmax'], name='Production of gas supply; {}; {}; '.format(k,t))
	g

	f = {}
	for l in L:
		for t in T:
			f[(l,t)] =  m.addVar(lb=-data['Lines'][l]['capacity'], ub=data['Lines'][l]['capacity'], name='Flow in line; {}; {}; '.format(l,t))
	f

	theta= {}
	for n in N:
		for t in T:
			theta[(n,t)] = m.addVar(lb=-np.pi, ub=np.pi, name='Voltage angle at node; {}; {}; '.format(n,t))
	theta

	q_in_plus= {}
	for z in Z:
		for t in T:
			q_in_plus[(z,t)] = m.addVar(lb=0, ub=gb.GRB.INFINITY, name='Inflow to pipeline; {}; {}; '.format(z,t))
	q_in_plus

	q_out_plus= {}
	for z in Z:
		for t in T:
			q_out_plus[(z,t)] = m.addVar(lb=0, ub=gb.GRB.INFINITY, name='Outflow from pipeline; {}; {}; '.format(z,t))
	q_out_plus

	q_in_minus= {}
	for z in Z:
		for t in T:
			q_in_minus[(z,t)] = m.addVar(lb=0, ub=gb.GRB.INFINITY, name='Inflow to pipeline; {}; {}; '.format(z,t))
	q_in_minus

	q_out_minus= {}
	for z in Z:
		for t in T:
			q_out_minus[(z,t)] = m.addVar(lb=0, ub=gb.GRB.INFINITY, name='Outflow from pipeline; {}; {}; '.format(z,t))
	q_out_minus

	q_plus= {}
	for z in Z:
		for t in T:
			q_plus[(z,t)] = m.addVar(lb=0, ub=gb.GRB.INFINITY, name='Flow in pipeline from m to u; {}; {}; '.format(z,t))
	q_plus

	q_minus= {}
	for z in Z:
		for t in T:
			q_minus[(z,t)] = m.addVar(lb=0, ub=gb.GRB.INFINITY, name='Flow in pipeline from u to m; {}; {}; '.format(z,t))
	q_minus

	h= {}
	for z in Z:
		for t in T:
			h[(z,t)] = m.addVar(lb=0, ub=gb.GRB.INFINITY, name='Linepack in pipeline; {}; {}; '.format(z,t))
	h

	q= {}
	for z in Z:
		for t in T:
			q[(z,t)] = m.addVar(lb=-gb.GRB.INFINITY, ub=gb.GRB.INFINITY, name='Flow over pipeline; {}; {}; '.format(z,t))
	q

	y= {}
	for z in Z:
		for t in T:
			y[(z,t)] = m.addVar(vtype=gb.GRB.BINARY, name='Direction of flow; {}; {}; '.format(z,t))
	y

	pr= {}
	for mz in MZ:
		for t in T:
			pr[(mz,t)] = m.addVar(lb=data['NGPr'][mz]['minPr'], ub=data['NGPr'][mz]['maxPr'], name='Pressure at node; {}; {}; '.format(mz,t))
	pr

	m.update()

	#Objective function
	m.setObjective(gb.quicksum(data['PP'][i]['C_E']*p[(i,t)] for i in C for t in T)  # Cost of running conventionals
		+ gb.quicksum(data['GasSupply'][k]['C_G'] * g[(k,t)] for k in K for t in T), gb.GRB.MINIMIZE)  # Cost of gas supply

	m.update()

	#Constraints:
	#Reference node constraint
	for t in T:
		ref_n = 'n1'
		m.addConstr(theta[(ref_n,t)],gb.GRB.EQUAL,0 ,name="reference node{}".format(t))

	for l in L:
		n_node = data['Lines'][l]['From']
		r_node = data['Lines'][l]['To']

		for t in T:
			m.addConstr(f[(l,t)],gb.GRB.EQUAL, (1/data['Lines'][l]['B'])*250 * (theta[(n_node,t)]-theta[(r_node,t)]) ,name="power flow over line{}{}".format(l,t))

	#Nodal power balance
	for n in N:
		A_I = units_in_node(data['PP'],n)
		A_J = units_in_node(data['Windfarm'],n)
		A_DE = units_in_node(data['ELDemand'],n)
		FromLines = units_in_node(l_n,n)        #Lines that start from node n
		ToLines = units_in_node(l_r,n)          #Lines that end in node n
		for t in T:
			m.addConstr(gb.quicksum(p[(i,t)] for i in A_I) + gb.quicksum(w[(j,t)] for j in A_J) - gb.quicksum(f[(l,t)] for l in FromLines) 
			+ gb.quicksum(f[(l,t)] for l in ToLines) ,gb.GRB.EQUAL, gb.quicksum(data['ELDemand'][d]['share']*data['Demand'][t]['EL'] for d in A_DE) ,name="power balance node{}{}".format(n,t))

	#Nodal gas balance
	for mz in MZ:
		A_K = units_in_node(data['GasSupply'],mz)
		A_DG = units_in_node(data['NGDemand'],mz)
		A_G = units_in_node(data['PP'],mz)
		FromPipelines = units_in_node(z_m,mz)        #Lines that start from node n
		ToPipelines = units_in_node(z_u,mz)          #Lines that end in node n
		for t in T:
			m.addConstr(gb.quicksum(g[(k,t)] for k in A_K)
			 - gb.quicksum(q_in_plus[(z,t)] for z in FromPipelines) 
			 + gb.quicksum(q_out_minus[(z,t)] for z in FromPipelines)
			 + gb.quicksum(q_out_plus[(z,t)] for z in ToPipelines)
			 - gb.quicksum(q_in_minus[(z,t)] for z in ToPipelines)
			 - gb.quicksum(data['PP'][i]['phi']*p[(i,t)] for i in A_G)
			 ,gb.GRB.EQUAL, gb.quicksum(data['NGDemand'][d]['share']*data['Demand'][t]['NG'] for d in A_DG) ,name="gasbalance node{}{}".format(m,t))

	for z in Z:
		m_node = data['KI_plus'][z]['From']
		u_node = data['KI_plus'][z]['To']

		for t in T:
			#Compressor
			if data['Pipelines'][z]['Gamma']!=1:
				m.addConstr(pr[(u_node,t)], gb.GRB.LESS_EQUAL, data['Pipelines'][z]['Gamma'] *pr[(m_node,t)] ,name="compressor{}{}".format(z,t))
			#Linear apprioximation around fixed pressure points
			for v in V:
				if data['KI_plus'][z][v]>0:
					m.addConstr(q_plus[(z,t)],gb.GRB.LESS_EQUAL,data['KI_plus'][z][v]*pr[(m_node,t)] - data['KO_plus'][z][v]*pr[(u_node,t)] + data['BigM']*(1-y[(z,t)]), name="flow plus from m to u {}{}".format(v,z,t))
				if data['KI_minus'][z][v]>0:
					m.addConstr(q_minus[(z,t)],gb.GRB.LESS_EQUAL,data['KI_minus'][z][v]*pr[(u_node,t)] - data['KO_minus'][z][v]*pr[(m_node,t)] + data['BigM']*y[(z,t)], name="flow minus from m to u {}{}".format(v,z,t))
			#Bidirectional flow
			m.addConstr(q[z,t],gb.GRB.EQUAL,q_plus[(z,t)]-q_minus[(z,t)],name="flow in pipeline {}{}".format(z,t))
			m.addConstr(q_plus[z,t],gb.GRB.LESS_EQUAL,data['BigM']*y[z,t],name="binary1 {}{}".format(z,t))
			m.addConstr(q_minus[z,t],gb.GRB.LESS_EQUAL,data['BigM']*(1-y[z,t]),name="binary2 {}{}".format(z,t))
			#Average flow
			m.addConstr(q_plus[(z,t)],gb.GRB.EQUAL,0.5*(q_in_plus[(z,t)]+q_out_plus[(z,t)]),name="average flow from m to u {}{}".format(z,t))
			m.addConstr(q_minus[(z,t)],gb.GRB.EQUAL,0.5*(q_in_minus[(z,t)]+q_out_minus[(z,t)]),name="average flow from u to m {}{}".format(z,t))
			#Linepack
			m.addConstr(h[(z,t)],gb.GRB.EQUAL,data['Pipelines'][z]['K_h']*0.5*(pr[(m_node,t)]+pr[(u_node,t)]),name="lineflow from m to u {}{}".format(z,t))
			#m.addConstr(q_in_plus[(z,t)],gb.GRB.EQUAL,q_out_plus[(z,t)],name="in and out flow from m to u {}{}".format(z,t))
			#m.addConstr(q_in_minus[(z,t)],gb.GRB.EQUAL,q_out_minus[(z,t)],name="in and out flow from u to m {}{}".format(z,t))
		#Linepack mass
		for t in T:
			#Initial linepack
			if t=='t1':
				m.addConstr(h[(z,t)],gb.GRB.EQUAL, data['Pipelines'][z]['H_ini']+ q_in_plus[z,t] - q_out_plus[z,t]+ q_in_minus[z,t] - q_out_minus[z,t],name="initial lineflow from m to u {}".format(z))
			#Linepack mass
			if t!='t1':
				previous = data['Demand'][t]['prev']
				m.addConstr(h[(z,t)],gb.GRB.EQUAL,h[(z,previous)] + q_in_plus[z,t] - q_out_plus[z,t] + q_in_minus[z,t] - q_out_minus[z,t],name="lineflow mass conservation from m to u {}{}".format(z,t))
			#Final linepack
			if t=='t24':
				m.addConstr(h[(z,t)],gb.GRB.GREATER_EQUAL, 1*data['Pipelines'][z]['H_ini'],name="final lineflow from m to u {}".format(z))

	m.Params.MIPGap= 0.0002
	m.optimize()
	#m.printAttr('X')
	return m.status

data=input()
status=MILP(data)
print(status)
